/*
 *  Copyright (C) 2006-2009  Ronald Blankendaal
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package org.dbgl.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;
import SevenZip.MyRandomAccessFile;
import SevenZip.Archive.IInArchive;
import SevenZip.Archive.SevenZipEntry;
import SevenZip.Archive.SevenZip.Handler;
import org.dbgl.gui.StreamGobbler;
import org.dbgl.loopy.iso9660.ISO9660FileEntry;
import org.dbgl.loopy.iso9660.ISO9660FileSystem;
import org.dbgl.model.DosboxVersion;
import org.dbgl.model.Profile;
import org.dbgl.model.Settings;
import org.dbgl.model.Template;
import org.dbgl.model.conf.CompositeConfiguration;
import org.dbgl.model.conf.Configuration;


public final class FileUtils {

    private static final int ZIP_BUFFER = 10240;
    
    private static final File DATA_DIR_FILE;
    private static final File DOSROOT_DIR_FILE;
    private static final File DOSBOX_DIR_FILE;
    private static final String TEMPLATES_DIR = "templates" + File.separatorChar;
    private static final String SETUP_CONF = "setup.conf";
    private static final String TEMPLATE_CONF = "template.conf";

    public static final String PROFILES_XML = "profiles.xml";
    public static final String DOSROOT_DIR = "dosroot" + File.separatorChar;
    public static final String CAPTURES_DIR = "captures" + File.separatorChar;
    public static final String PROFILES_DIR = "profiles" + File.separatorChar;
    public static final String EXPORT_DIR = "export" + File.separatorChar;
    public static final String XSL_DIR = "xsl" + File.separatorChar;
    public static final String XSL_EXT = ".xsl";
    public static final String XML_EXT = ".xml";
    public static final String CONF_EXT = ".conf";
    public static final String GAMEPACKARCHIVE_EXT = ".dbgl.zip";
    public static final String[] CDIMAGES = { ".iso", ".cue", ".bin" };
    public static final String[] EXECUTABLES = { ".exe", ".com", ".bat" };
    public static final String[] ARCHIVES = { ".zip", ".7z" };
    public static final String[] BOOTERIMAGES = { ".cp2", ".dcf", ".img", ".jrc", ".td0" };
    public static final String[] PICTURES = { ".png", ".gif", ".jpg", ".tif", ".tiff", ".ico", ".bmp" };

    public static final String DOSBOX_CONF = "dosbox.conf";
    public static final String CNF_FILTER = "*.conf;*.CONF";
    public static final String EXE_FILTER = "*.com;*.COM;*.exe;*.EXE;*.bat;*.BAT";
    public static final String ARC_FILTER = "*.zip;*.ZIP;*.7z;*.7Z";
    public static final String DBGLZIP_FILTER = "*.dbgl.zip;*.DBGL.ZIP";
    public static final String BTR_FILTER = "*.cp2;*.CP2;*.dcf;*.DCF;*.img;*.IMG;*.jrc;*.JRC;*.td0;*.TD0";
    public static final String CDI_FILTER = "*.iso;*.ISO;*.cue;*.CUE";
    public static final String ALL_FILTER = "*";
    
    public static final String INVALID_FILENAME_CHARS_REGEXP = "[^a-zA-Z_0-9()]";


    static {
        final Settings settings = Settings.getInstance();
        final String DATADIR = getCanonicalPath(settings.getValue("directory", "data"));
        DATA_DIR_FILE = new File(DATADIR);
        DOSROOT_DIR_FILE = new File(DATADIR, DOSROOT_DIR);
        final String DOSBOXDIR = getCanonicalPath(settings.getValue("directory", "dosbox"));
        DOSBOX_DIR_FILE = new File(DOSBOXDIR);

        createDosrootIfNecessary();
    }

    public static String getCanonicalPath(final String path) {
        String canPath = path;
        if ((PlatformUtils.IS_LINUX || PlatformUtils.IS_OSX) &&
                (canPath.startsWith("~/") || (canPath.length() == 1 && canPath.charAt(0) == '~'))) {
            //Linux and OSX have ~/ for homedirectory. Allow single ~ as well
            canPath = canPath.replaceAll("^~", System.getProperty("user.home"));
        }
        try {
            return new File(canPath).getCanonicalPath();
        } catch (IOException e) {
            return new File(canPath).getAbsolutePath();
        }
    }

    public static boolean isExistingFile(final File file) {
        return file.isFile() && file.exists();
    }

    private static void copy(final File source, final File dest) {
        FileChannel inStream = null, outStream = null;
        try {          
            inStream = new FileInputStream(source).getChannel();
            outStream = new FileOutputStream(dest).getChannel();
            inStream.transferTo(0, inStream.size(), outStream);
        } catch (IOException e) {
            System.err.println(Settings.getInstance().msg("general.error.copyfile", new Object[] {source, dest}));
        } finally {
            if (inStream != null) { 
                try {
                    inStream.close();
                } catch (IOException ex) {
                    System.err.println(Settings.getInstance().msg("general.error.closefileinputstream", new Object[] {source}));
                }
            }
            if (outStream != null) {
                try {
                    outStream.close();
                } catch (IOException ex) {
                    System.err.println(Settings.getInstance().msg("general.error.closefileoutputstream", new Object[] {source}));
                }
            }
        }
    }
    
    private static List<String> initCommands(final DosboxVersion dbversion, boolean forceDBConf) {
    	List<String> execCommands = new ArrayList<String>();
    	if (dbversion.isUsingCurses() && PlatformUtils.IS_WINDOWS) {
    		execCommands.add("rundll32");
    		execCommands.add("SHELL32.DLL,ShellExec_RunDLL");
        }
    	execCommands.add(dbversion.getCanonicalExecutable().getPath());
    	if ((dbversion.isMultiConfig() && isReadableFile(dbversion.getCanonicalConfFile())) || forceDBConf) {
            // selected default dosbox config file
    		execCommands.add("-conf");
    		execCommands.add(dbversion.getCanonicalConfFile().getPath()); 
        }
    	return execCommands;
    }
    
    private static void postCommands(final DosboxVersion dbversion, List<String> execCommands) {
    	if (dbversion.getParameters().length() > 0) {
            for (String p: dbversion.getParameters().split(" ")) {
            	execCommands.add(p);
            }
        }
    	if (Settings.getInstance().getBooleanValue("dosbox", "hideconsole")) {
            execCommands.add("-noconsole");
        }
    }

    private static void executeCommand(final List<String> execCommands, final File cwd,
    		Map<String, String> env, final boolean waitFor) throws IOException {
        StringBuffer cmd = new StringBuffer();
        try {
            String[] rtCmds = new String[execCommands.size()];
            for (int i = 0; i < execCommands.size(); i++) {
                rtCmds[i] = execCommands.get(i);
                cmd.append(rtCmds[i]).append(' ');
            }
            System.out.print(cmd);
            File dir = (cwd == null)? DOSROOT_DIR_FILE: cwd;
            ProcessBuilder pb = new ProcessBuilder(rtCmds);
            pb.directory(dir);
            Map<String, String> environment = pb.environment();
            if (env != null) {
            	environment.putAll(env);
            	System.out.print(env);
            }
            System.out.println();
            Process proc = pb.start();
            StreamGobbler errorGobbler = new StreamGobbler(proc.getErrorStream(), "DOSBox stderr");            
            StreamGobbler outputGobbler = new StreamGobbler(proc.getInputStream(), "DOSBox stdout");
            outputGobbler.start();
            errorGobbler.start();
            if (waitFor) {
                try {	proc.waitFor();} catch (InterruptedException e) {}
            }
        } catch (IOException e) {
            e.printStackTrace();
            throw new IOException(Settings.getInstance().msg("general.error.startdosbox", new Object[] {cmd}));
        }
    }

    private static void doRunDosbox(final DosboxVersion dbversion, final String[] parameters,
            final boolean forceDBConf, final File cwd, final Map<String, String> env, final boolean waitFor) throws IOException {
        List<String> commandItems = initCommands(dbversion, forceDBConf);
        commandItems.addAll(Arrays.asList(parameters));
        postCommands(dbversion, commandItems);
        executeCommand(commandItems, cwd, env, waitFor);
    }
  
    private static File canonicalTo(final File base, final String path) {
        File file = new File(path);
        if (file.isAbsolute()) {
        	try {
                return file.getCanonicalFile();
            } catch (IOException e) {
                return file.getAbsoluteFile();
            }
        } else {
            try {
                return new File(base, file.getPath()).getCanonicalFile();
            } catch (IOException e) {
                return new File(base, file.getPath()).getAbsoluteFile();
            }
        }
    }

    public static boolean areRelated(final File parent, final File child) {
        File remainder = child.getParentFile();
        while (remainder != null) {
            if (parent.equals(remainder)) {
                return true;
            }
            remainder = remainder.getParentFile();
        }
        return false;
    }
  
    public static String getDosRoot() {
        return DOSROOT_DIR_FILE.getPath();
    }

    public static File makeRelativeTo(final File file, final File basePath) {
        if (!file.isAbsolute()) {
            return file;
        }
        if (file.equals(basePath)) {
            return new File(".");
        }
        File remainder = new File(file.getName());
        File parent = file.getParentFile();
        while (parent != null) {
            if (parent.equals(basePath)) {
                return remainder;
            }
            remainder = new File(parent.getName(), remainder.getPath());
            parent = parent.getParentFile();
        }
        return file;
    }
  
    public static File makeRelativeToData(final File file) {
        return makeRelativeTo(file, DATA_DIR_FILE);
    }

    public static File makeRelativeToDosroot(final File file) {
        return makeRelativeTo(file, DOSROOT_DIR_FILE);
    }

    public static File makeRelativeToDosbox(final File file) {
        return makeRelativeTo(file, DOSBOX_DIR_FILE);
    }

    public static File canonicalToData(final String path) {
        return canonicalTo(DATA_DIR_FILE, path);
    }

    public static File canonicalToDosbox(final String path) {
        return canonicalTo(DOSBOX_DIR_FILE, path);
    }

    public static File canonicalToDosroot(final String path) {
        return canonicalTo(DOSROOT_DIR_FILE, path);
    }

    public static String sanitizeToDosroot(final String path) {
        return makeRelativeToDosroot(canonicalToDosroot(path)).getPath();
    }

    public static File prefixAndSanitizeToDosroot(final File basePath, final File file) {
    	if (!file.isAbsolute())
    		return makeRelativeToDosroot(canonicalToDosroot(new File(basePath, file.getPath()).getPath()));
   		return file;
    }

    public static File constructCanonicalDBConfLocation(final String path) {
        return canonicalToDosbox(new File(path, DOSBOX_CONF).getPath());
    }

    public static File constructCanonicalDBExeLocation(final String path) {
        return canonicalToDosbox(new File(path, PlatformUtils.DB_EXECUTABLE).getPath());
    }

    public static String constructCapturesDir(final int profileId) {
        return CAPTURES_DIR + profileId;
    }

    public static String constructRelativeCapturesDir(final int profileId) {
        return ".." + File.separatorChar + constructCapturesDir(profileId);
    }

    public static File constructCanonicalTemplateFileLocation(final int templateId) {
        return canonicalToData(TEMPLATES_DIR + templateId + CONF_EXT);
    }

    public static String constructUniqueConfigFileString(final int profileId, final String profileTitle, final Configuration gameConf) {
        Settings set = Settings.getInstance();
        File candidate = null;
        int candidateNr = 1;
        boolean unique = false;
        while (!unique) {
            File candidatePath;
            if (set.getIntValue("profiledefaults", "confpath") == 1) {
                candidatePath = gameConf.getCanonicalMainDir();
            } else {
                candidatePath = new File(PROFILES_DIR);
            }

            String candidateFile;
            if (set.getIntValue("profiledefaults", "conffile") == 1) {
                String candidateTitle = profileTitle;
                if (candidateNr > 1) {
                    candidateTitle += "(" + candidateNr + ")";
                }
                candidateFile = candidateTitle.replaceAll(INVALID_FILENAME_CHARS_REGEXP, "") + CONF_EXT;
            } else {
                if (candidateNr > 1) {
                    return null; // id files should be unique, else give up
                }
                candidateFile = profileId + CONF_EXT;
            }
            candidateNr++;
            candidate = new File(candidatePath, candidateFile);
            unique = !isExistingFile(candidate);
        }
        return candidate.getPath();
    }

    public static void doRunDosbox(final DosboxVersion dbversion, final Map<String, String> env) throws IOException {
        doRunDosbox(dbversion, new String[] {}, true, null, env, false);
    }

    public static void doCreateDosboxConf(final DosboxVersion dbversion) throws IOException {
        doRunDosbox(dbversion, new String[] {"-c", "config -writeconf " + DOSBOX_CONF, "-c", "exit"}, false, dbversion.getCanonicalExecutable().getParentFile(), null, true);
    }
  
    public static void doRunProfile(final Profile prof, final List<DosboxVersion> dbversions, final Map<String, String> env,
            final boolean setup) throws IOException {
        DosboxVersion dbv = dbversions.get(DosboxVersion.findById(dbversions, prof.getDbversionId()));
        doRunProfile(prof, dbv, env, setup);
    }

    public static void doRunTemplate(final Template template, final List<DosboxVersion> dbversions, final Map<String, String> env) throws IOException {
        DosboxVersion dbv = dbversions.get(DosboxVersion.findById(dbversions, template.getDbversionId()));
        doRunTemplate(template, dbv, env);
    }

    public static void doRunProfile(final Profile prof, final DosboxVersion dbversion, Map<String, String> env,
            final boolean setup) throws IOException {
        File file = prof.getCanonicalConfFile();
        if (setup && prof.hasSetup()) {
        	CompositeConfiguration cconf = new CompositeConfiguration(dbversion, prof, false, null, System.out);
        	Configuration conf = cconf.getCompositeConf();
        	conf.setAutoexecSettings(prof.getSetup(), prof.getSetupParameters());
            file = canonicalToData(PROFILES_DIR + SETUP_CONF);
            cconf.save(file, null, System.out);
        }
        doRunDosbox(dbversion, new String[] {"-conf", file.getPath()}, false, null, env, false);
    }
    
    public static void doRunTemplate(final Template template, final DosboxVersion dbversion, Map<String, String> env) throws IOException {
        File file = constructCanonicalTemplateFileLocation(template.getId());
        CompositeConfiguration cconf = new CompositeConfiguration(dbversion, template, false, null, System.out);
    	Configuration conf = cconf.getCompositeConf();
    	conf.setAutoexecSettings(false, "", "", "", "", "");
    	conf.setAutoexecSettings(false);
        file = canonicalToData(PROFILES_DIR + TEMPLATE_CONF);
        cconf.save(file, null, System.out);
        doRunDosbox(dbversion, new String[] {"-conf", file.getPath()}, false, null, env, false);
    }
  
    public static void createDir(final File dir) {
        if (!dir.exists() && !dir.mkdirs()) {
            System.err.println(Settings.getInstance().msg("general.error.createdir", new Object[] {dir}));
        }
    }

    public static void copyFiles(final File srcDir, final File dstDir) {
        File[] srcFiles = srcDir.listFiles();
        if (srcFiles != null) {
            for (File src: srcFiles) {
                if (src.isFile()) {
                    File dst = new File(dstDir, src.getName());
                    copy(src, dst);
                    dst.setLastModified(src.lastModified());
                }
            }
        }
    }

    public static void removeFile(final File file) {
        if (! (file.isFile() && file.delete()) ) {
            System.err.println(Settings.getInstance().msg("general.error.deletefile", new Object[] {file.getPath()}));
        }      
    }

    public static void removeFilesInDirAndDir(final File dir) {
        File[] files = dir.listFiles();
        if (files != null) {
            for (File file: files) {
                if (file.isDirectory()) {
                    System.err.println(Settings.getInstance().msg("general.error.dirtobedeletedcontainsdir", new Object[] {dir.getPath()}));
                    return;
                }
            }
            for (File file: files) {
                if (!file.delete()) {
                    System.err.println(Settings.getInstance().msg("general.error.deletefile", new Object[] {file}));
                }
            }
        }
        if (! (dir.isDirectory() && dir.delete()) ) {
            System.err.println(Settings.getInstance().msg("general.error.deletedir", new Object[] {dir.getPath()}));
        }
    }
  
    public static boolean isReadableFile(final File file) {
        return file.isFile() && file.canRead();
    }

    public static void createDosrootIfNecessary() {
        if (!DOSROOT_DIR_FILE.isDirectory()) {
            System.out.println(Settings.getInstance().msg("general.notice.createdir", new Object[] {DOSROOT_DIR_FILE.getPath()}));
            if (!DOSROOT_DIR_FILE.mkdirs()) {
                System.err.println(Settings.getInstance().msg("general.error.createdir", new Object[] {DOSROOT_DIR_FILE.getPath()}));
            }
        }
    }

    public static boolean isExecutable(final String filename) {
        for (String ext: EXECUTABLES) {
            if (filename.toLowerCase().endsWith(ext)) {
                return true;
            }
        }
        return false;
    }
  
    public static boolean isArchive(final String filename) {
        for (String ext: ARCHIVES) {
            if (filename.toLowerCase().endsWith(ext)) {
                return true;
            }
        }
        return false;
    }

    public static boolean isPhysFS(final String mountPath) {
        for (String ext: ARCHIVES) {
            if (mountPath.toLowerCase().endsWith(ext + ':' + File.separatorChar)) {
                return true;
            }
        }
        return false;
    }
  
    public static int containsPhysFS(final String mountPath) {
        for (String ext: ARCHIVES) {
            int idx = mountPath.toLowerCase().indexOf(ext + ':' + File.separatorChar);
            if (idx != -1) {
                return idx + ext.length();
            }
        }
        return -1;
    }

    public static boolean isBooterImage(final String filename) {
        for (String ext: BOOTERIMAGES) {
            if (filename.toLowerCase().endsWith(ext)) {
                return true;
            }
        }
        return false;
    }

    public static boolean isConfFile(final String filename) {
        return filename.toLowerCase().endsWith(CONF_EXT);
    }
    
    public static boolean isGamePackArchiveFile(final String filename) {
        return filename.toLowerCase().endsWith(GAMEPACKARCHIVE_EXT);
    }

    public static boolean isIsoFile(final String filename) {
        for (String ext: CDIMAGES) {
            if (filename.toLowerCase().endsWith(ext)) {
                return true;
            }
        }
        return false;
    }
    
    public static int containsIso(final String mountPath) {
        for (String ext: CDIMAGES) {
            int idx = mountPath.toLowerCase().indexOf(ext + File.separatorChar);
            if (idx != -1) {
                return idx + ext.length();
            }
        }
        return -1;
    }
    
    public static boolean isPicture(final String filename) {
        for (String ext: PICTURES) {
            if (filename.toLowerCase().endsWith(ext)) {
                return true;
            }
        }
        return false;
    }
    
    public static File getCanMainFile(final File file) {
        String f = file.getPath();
        int isoIdx = containsIso(f);
        int pfsIdx = containsPhysFS(f);
        if (isoIdx != -1) {
            return canonicalToDosroot(f.substring(0, isoIdx));
        } else if (pfsIdx != -1) {
            return canonicalToDosroot(f.substring(0, pfsIdx));
        }
        return canonicalToDosroot(file.getPath());
    }

    public static String[] getExecutablesInZipOrIso(final String archive) throws IOException {
        List<String> result = new ArrayList<String>();
        File arcFile = new File(archive);

        if (archive.toLowerCase().endsWith(ARCHIVES[0])) { // zip
            ZipFile zfile = new ZipFile(arcFile);
            for (Enumeration<? extends ZipEntry> entries = zfile.entries(); entries.hasMoreElements();) {
                ZipEntry entry = entries.nextElement();
                String name = entry.getName();
                if (!entry.isDirectory() && isExecutable(name)) {
                    result.add(PlatformUtils.archiveToNativePath(name));
                }
            }
            zfile.close();
        } else if (archive.toLowerCase().endsWith(ARCHIVES[1])) { // 7-zip
            MyRandomAccessFile istream = new MyRandomAccessFile(archive, "r");
            IInArchive zArchive = new Handler();
            if (zArchive.Open(istream) != 0) {
                throw new IOException(Settings.getInstance().msg("general.error.opensevenzip", new Object[] {archive}));
            }
            for (int i = 0; i < zArchive.size(); i++) {
                SevenZipEntry entry = zArchive.getEntry(i);
                String name = entry.getName();
                if (!entry.isDirectory() && isExecutable(name)) {
                    result.add(PlatformUtils.archiveToNativePath(name));
                }
            }
            zArchive.close();
        } else if (isIsoFile(archive)) {
            ISO9660FileSystem iso = new ISO9660FileSystem(new File(archive));
            for (Enumeration<ISO9660FileEntry> entries = iso.getEntries(); entries.hasMoreElements();) {
                ISO9660FileEntry entry = entries.nextElement();
                String name = entry.getPath();
                if (!entry.isDirectory() && isExecutable(name)) {
                    result.add(PlatformUtils.archiveToNativePath(name));
                }
            }
            iso.close();
        }

        Collections.sort(result, new Comparator<String>() {
            public int compare(final String string1, final String string2) {
                int count1 = StringUtils.countCharacters(string1, '\\');
                int count2 = StringUtils.countCharacters(string2, '\\');
                if (count1 == count2) {
                    return string1.compareTo(string2);
                } else {
                    return count1 - count2;
                }
            }
        });

        return result.toArray(new String[result.size()]);
    }
    
    public static void zipEntry(final File orgFile, final File fileEntry, final ZipOutputStream zos) throws IOException {
        ZipEntry anEntry = new ZipEntry(PlatformUtils.toArchivePath(fileEntry, orgFile.isDirectory()));
        anEntry.setTime(orgFile.lastModified());
        if (orgFile.isFile() && !orgFile.canWrite()) anEntry.setExtra(new byte[] {1});
        zos.putNextEntry(anEntry);

        if (orgFile.isFile()) {
	        byte[] readBuffer = new byte[ZIP_BUFFER];
	        int bytes = 0;
	        FileInputStream is = new FileInputStream(orgFile);
	        while ((bytes = is.read(readBuffer)) != -1)
	            zos.write(readBuffer, 0, bytes);
	        is.close();
        }
        zos.closeEntry();
    }
    
    public static void zipDir(final File dirToZip, final ZipOutputStream zos, final File dstDirInZip) throws IOException {
        zipDir(dirToZip, zos, dirToZip, dstDirInZip);
    }

    private static void zipDir(final File dirToZip, final ZipOutputStream zos, final File zipRootDir, final File dstDirInZip) throws IOException {
        String[] dirList = dirToZip.list();
        if (dirList == null) throw new IOException(Settings.getInstance().msg("general.error.opendir", new Object[] {dirToZip}));
        for (int i = 0; i < dirList.length; i++) {
            File f = new File(dirToZip, dirList[i]);
            zipEntry(f, new File(dstDirInZip, makeRelativeTo(f, zipRootDir).getPath()), zos);
            if (f.isDirectory()) {
                zipDir(f, zos, zipRootDir, dstDirInZip);
            }
        }
    }

    public static long extractZipSizeInBytes(final File archive, final File dirToBeExtracted) throws IOException {
    	long bytes = 0;
    	ZipFile zf = new ZipFile(archive);
    	for (Enumeration<? extends ZipEntry> entries = zf.entries(); entries.hasMoreElements();) {
    		ZipEntry entry = entries.nextElement();
    		if (areRelated(dirToBeExtracted, new File(entry.getName())))
    			bytes += entry.getSize();
    	}
    	zf.close();
    	return bytes;
    }

    public static void extractZip(final File archive, final File dirToBeExtracted, final File destDir, final ProgressNotifyable prog) throws IOException {
    	ZipFile zf = new ZipFile(archive);
    	for (Enumeration<? extends ZipEntry> entries = zf.entries(); entries.hasMoreElements();) {
    		ZipEntry entry = entries.nextElement();
    		if (areRelated(dirToBeExtracted, new File(entry.getName())))
    			extractEntry(zf, dirToBeExtracted, entry, destDir, prog);
    	}
    	zf.close();
    }

    private static void extractEntry(final ZipFile zf, final File dirToBeExtracted, final ZipEntry entry, final File destDir, final ProgressNotifyable prog) throws IOException {
    	File destFile = canonicalTo(destDir, strip(new File(entry.getName()), dirToBeExtracted).getPath());
    	if (destFile.exists()) throw new IOException(Settings.getInstance().msg("general.error.filetobeextractedexists", new Object[] {destFile}));
    	if (entry.isDirectory()) {
    		if (!destFile.exists()) createDir(destFile);
    	} else {
    		if (destFile.getParentFile() != null) createDir(destFile.getParentFile());
        	FileOutputStream fos = new FileOutputStream(destFile);
        	InputStream is = zf.getInputStream(entry);
        	byte[] readBuffer = new byte[ZIP_BUFFER];
        	int bytesIn;
        	while ((bytesIn = is.read(readBuffer)) != -1)
        		fos.write(readBuffer, 0, bytesIn);
        	fos.flush();
        	fos.close();
        	is.close();
        	byte[] extra = entry.getExtra();
        	if ((extra != null) && (extra[0] == 1)) destFile.setReadOnly();
    	}
    	destFile.setLastModified(entry.getTime());
    	prog.notifyOfProgress(entry.getSize());
    }

    private static File strip(final File file, final File basePath) {
        if (file.equals(basePath)) return new File(".");
        File remainder = new File(file.getName());
        File parent = file.getParentFile();
        while (parent != null) {
            if (parent.equals(basePath)) return remainder;
            remainder = new File(parent.getName(), remainder.getPath());
            parent = parent.getParentFile();
        }
        return file;
    }

    public static String getUrlFromFile(final File file) throws MalformedURLException {
        return file.toURI().toURL().toString();
    }
}
